cdef class Scatter2D(AnalysisObject):
    """
    2D scatter plot, i.e. a collection of Point2D objects with positions and errors.

    Constructor calling idioms:

    Scatter2D(path="", title="")
      Create a new empty scatter, with optional path and title.

    Scatter2D(points, path="", title=""):
      Create a new empty scatter from an iterable of points, with optional path
      and title.

    TODO: more documentation!
    """

    cdef inline c.Scatter2D* _Scatter2D(self) except NULL:
        return <c.Scatter2D*> self.ptr()

    def __init__(self, *args, **kwargs):
        util.try_loop([self.__init_2, self.__init_3], *args, **kwargs)

    def __init_2(self, char* path="", char* title=""):
        util.set_owned_ptr(self, new c.Scatter2D(string(path), string(title)))

    def __init_3(self, points, char* path="", char* title=""):
        self.__init_2(path, title)
        self.addPoints(points)


    def __len__(self):
        return self._Scatter2D().numPoints()

    def __getitem__(self, py_ix):
        cdef size_t i = util.pythonic_index(py_ix, self._Scatter2D().numPoints())
        return util.new_borrowed_cls(Point2D, & self._Scatter2D().point(i), self)

    def __repr__(self):
        return "<%s '%s' %d points>" % \
               (self.__class__.__name__, self.path, len(self.points))


    def clone(self):
        """() -> Scatter2D.
        Clone this Scatter2D."""
        return util.new_owned_cls(Scatter2D, self._Scatter2D().newclone())


    @property
    def numPoints(self):
        """() -> int
        Number of points in this scatter."""
        return self._Scatter2D().numPoints()

    @property
    def points(self):
        """Access the ordered list of points."""
        return list(self)

    def addPoint(self, *args, **kwargs):
        """Add a new point."""
        try:
            self.__addPoint_point(*args, **kwargs)
        except TypeError:
            self.__addPoint_explicit(*args, **kwargs)

    def __addPoint_explicit(self, x, y, xerrs=0, yerrs=0):
        self.__addPoint_point(Point2D(x, y, xerrs, yerrs))

    def __addPoint_point(self, Point2D p):
        self._Scatter2D().addPoint(p._Point2D()[0])

    def addPoints(self, iterable):
        """Add several new points."""
        for row in iterable:
            self.addPoint(*row)

    def combineWith(self, others):
        """Try to add points from other Scatter2Ds into this one."""
        cdef Scatter2D other
        try:
            # Can we type it as a Scatter2D?
            other = others
        except TypeError:
            # Could be an iterable...
            for other in others:
                self._Scatter2D().combineWith(deref(other._Scatter2D()))
        else:
            self._Scatter2D().combineWith(deref(other._Scatter2D()))


    def scale(self, ax, ay):
        """(float,float) -> None
        Scale the x values and errors of the points in this scatter by factors ax, ay."""
        self._Scatter2D().scale(ax, ay)

    def transformX(self, f):
        """(fn) -> None
        Transform the x values and errors of the points in this scatter by function f."""
        # Import ctypes here (rather than at module import) so that Python 2.4
        # users don't need ctypes to use the rest of the library.
        import ctypes
        try:
            callback = ctypes.CFUNCTYPE(ctypes.c_double, ctypes.c_double)(f)
        except:
            raise RuntimeError("Callback is not of type (double) -> double")
        fptr = (<c.dbl_dbl_fptr*><size_t>ctypes.addressof(callback))[0]
        c.Scatter2D_transformX(deref(self._Scatter2D()), fptr)

    def transformY(self, f):
        """(fn) -> None
        Transform the y values and errors of the points in this scatter by function f."""
        # Import ctypes here (rather than at module import) so that Python 2.4
        # users don't need ctypes to use the rest of the library.
        import ctypes
        try:
            callback = ctypes.CFUNCTYPE(ctypes.c_double, ctypes.c_double)(f)
        except:
            raise RuntimeError("Callback is not of type (double) -> double")
        fptr = (<c.dbl_dbl_fptr*><size_t>ctypes.addressof(callback))[0]
        c.Scatter2D_transformY(deref(self._Scatter2D()), fptr)


    def __add__(Scatter2D self, Scatter2D other):
        h = Scatter2D()
        util.set_owned_ptr(h, c.Scatter2D_add_Scatter2D(self._Scatter2D(), other._Scatter2D()))
        return h

    def __sub__(Scatter2D self, Scatter2D other):
        h = Scatter2D()
        util.set_owned_ptr(h, c.Scatter2D_sub_Scatter2D(self._Scatter2D(), other._Scatter2D()))
        return h
